"""
Time series mixin for moving sum, average, product
"""


#[

from __future__ import annotations

from collections.abc import Iterable
from typing import Callable
import numpy as _np
import textwrap as _tw
import documark as _dm

from ._functionalize import FUNC_STRING

#]


#-------------------------------------------------------------------------------
# Mixin methods
#-------------------------------------------------------------------------------


class Mixin:
    #[

    @_dm.reference(
        category=None,
        call_name="Moving window calculations",
        call_name_is_code=False,
        priority=20,
    )
    def moving_window(
        self,
        func: Callable,
        window: int | None = None,
    ) -> None:
        r"""
    ················································································


    Overview of moving window functions:

    | Function | Description
    |----------|-------------
    | `mov_sum` | Moving sum, $y_t = \sum_{i=0}^{k-1} x_{t-i}$
    | `mov_avg` | Moving average, $y_t = \frac{1}{k} \sum_{i=0}^{k-1} x_{t-i}$
    | `mov_mean` | Same as `mov_avg`
    | `mov_prod` | Moving product, $y_t = \prod_{i=0}^{k-1} x_{t-i}$

    where

    * $k$ is the window length determined by the `window` input argument
    (note that $k$ above is a positive integer while `window` needs to be
    entered as a negative integer).



    ### Function for creating new Series objects ###

    ```
    new = irispie.mov_sum(self, window=None, )
    new = irispie.mov_avg(self, window=None, )
    new = irispie.mov_mean(self, window=None, )
    new = irispie.mov_prod(self, window=None, )
    ```


    ### Methods for changing existing Series objects in-place ###


    ```
    self.mov_sum(window=None, )
    self.mov_avg(window=None, )
    self.mov_mean(window=None, )
    self.mov_prod(window=None, )
    ```


    ### Input arguments ###


    ???+ input "self"
    Time series on whose data a moving function is calculated (see the
    overview table above).

    ???+ input "window"
    A negative interger determining the number of observations to include
    in the moving window, counting from the current time period backwards
    (the minus sign is a convention to indicate that the window goes
    backwards in time). If `window=None` (or not specified), the default
    window is derived from the time series frequency:

    | Date frequency | Default window |
    |----------------|---------------:|
    | `YEARLY`       |             –1 |
    | `QUARTERLY`    |             –4 |
    | `MONTHLY`      |            –12 |
    | `WEEKLY`       |            –52 |
    | `DAILY`        |           –365 |
    | Otherwise      |             –4 |

    Note that the default windows for `DAILY` and `WEEKLY` frequencies are
    fixed numbers and do not depend on the actual number of days in a
    calendar year or the actual number of weeks in a calendar year.


    ### Returns ###


    ???+ returns "new"
    New time series with data calculated as a moving window function of the
    original data.

    ???+ returns "self"
    Time series with data replaced by the moving window function of the
    original data.

    ················································································
        """
        window = (
            window if window is not None
            else _get_default_moving_window(self, )
        )
        window_length = -window
        data = _np.pad(
            self.data,
            pad_width=((window_length-1, 0), (0, 0)),
            mode="constant",
            constant_values=_np.nan,
        )
        data_windows =_np.lib.stride_tricks.sliding_window_view(
            data,
            window_shape=window_length,
            axis=0,
        )
        new_data = func(data_windows, axis=2, )
        self._replace_data(new_data, )

    @_dm.reference(category="moving", )
    def mov_sum(self, window: int | None = None, ) -> None:
        r"""
    ................................................................................

    ==Moving sum==


    See documentation of [moving window calculations](#moving-window-calculations) or
    `help(irispie.Series.moving_window)`.

    ................................................................................
        """
        self.moving_window(_np.sum, window=window, )

    @_dm.reference(category="moving", )
    def mov_avg(self, window: int | None = None, ) -> None:
        r"""
    ................................................................................

    ==Moving average==


    See documentation of [moving window calculations](#moving-window-calculations) or
    `help(irispie.Series.moving_window)`.

    ................................................................................
        """
        self.moving_window(_np.mean, window=window, )

    @_dm.reference(category="moving", )
    def mov_mean(self, window: int | None = None, ) -> None:
        r"""
    ................................................................................

    ==Moving average==


    See documentation of [moving window calculations](#moving-window-calculations) or
    `help(irispie.Series.moving_window)`.

    ................................................................................
        """
        self.moving_window(_np.mean, window=window, )

    @_dm.reference(category="moving", )
    def mov_prod(self, window: int | None = None, ) -> None:
        r"""
................................................................................

==Moving product==


See documentation of [moving window calculations](#moving-window-calculations) or
`help(irispie.Series.moving_window)`.

................................................................................
        """
        self.moving_window(_np.prod, window=window, )


    #]


#-------------------------------------------------------------------------------
# Functional forms
#-------------------------------------------------------------------------------


_functional_forms = {
    "moving_window",
    "mov_sum",
    "mov_avg",
    "mov_mean",
    "mov_prod",
}

for n in _functional_forms:
    code = FUNC_STRING.format(n=n, )
    exec(_tw.dedent(code, ), globals(), )

__all__ = tuple(_functional_forms)


#-------------------------------------------------------------------------------


def _get_default_moving_window(self, ) -> int:
    r"""
    Derive default moving window from time series frequency
    """
    return (
        -self.frequency.value
        if self.frequency is not None and self.frequency.value > 0
        else -4
    )




r"""
Create a namespace for functional forms of Series methods
"""


#[

from __future__ import annotations

#]

from ._extrapolate import *
from ._extrapolate import __all__ as _all_extrapolate

from ._temporal import *
from ._temporal import __all__ as _all_temporal

from ._conversions import *
from ._conversions import __all__ as _all_conversions

from ._filling import __all__ as _all_filling
from ._filling import *

from ._lays import __all__ as _all_lays
from ._lays import *

from ._hp import __all__ as _all_hp
from ._hp import *

from ._x13 import __all__ as _all_x13
from ._x13 import *

from ._timing import *
from ._timing import __all__ as _all_timing

from ._moving import *
from ._moving import __all__ as _all_moving

from ._ell_one import __all__ as _all_ell_one
from ._ell_one import *

from ._statistics import *
from ._statistics import __all__ as _statistics_all
from ._statistics import functional_form_context as _statistics_functional_form_context

from ._elementwise import *
from ._elementwise import __all__ as _elementwise_all
from ._elementwise import functional_form_context as _elementwise_functional_form_context


__all__ = (
    *_all_extrapolate,
    *_all_temporal,
    *_all_conversions,
    *_all_filling,
    *_all_lays,
    *_all_hp,
    *_all_x13,
    *_all_timing,
    *_all_moving,
    *_all_ell_one,
    *_statistics_all,
    *_elementwise_all,
)


functional_form_context = {}
functional_form_context.update(_elementwise_functional_form_context, )
functional_form_context.update(_statistics_functional_form_context, )




"""
Time series indexing inlay
"""


#[

from __future__ import annotations

import documark as _dm

from .. import dates as _dates
from ._categories import CATEGORIES

#]


#-------------------------------------------------------------------------------
# Mixin methods
#-------------------------------------------------------------------------------


class Mixin:
    #[

    @_dm.reference(
        category=None,
        call_name=CATEGORIES["indexing"],
        call_name_is_code=False,
        priority=30,
    )
    def indexing(self, /, ) -> None:
        r"""
    ················································································

    Time `Series` objects can be indexed in four ways (note the square versus round
    brackets):

    | Indexing                                           | Description
    |----------------------------------------------------|-------------
    | `self[shift]`                                      | Time shift
    | `self[dates]`, `self[dates, variants]`             | Data extraction
    | `self[dates] = ...`, `self[dates, variants] = ...` | Data assignment
    | `self(dates)`, `self(dates, variants)`             | Time `Series` recreation


    ### Time shift ###

    ```
    self[shift]
    ```

    Time shift is done by passing an integer to the `self[shift]` or
    `self[shift]` indexing. The time shift syntax returns a new copy of the
    original series, with the time periods shifted by `shift`.


    ### Data extractation ###

    ```
    self[dates]
    self[dates, variants]
    ```

    The `dates` is a `Dater` or a tuple of `Daters` or a time `Span` object,
    and `variants` is an integer or a tuple of integers or a `slice` object
    specifying the variants. The data extraction syntax returns a
    two-dimensional `numpy` array, with the time dimension running along the
    rows and the variant dimension running along the columns.


    ### Data assignment ###

    ```
    self[dates] = ...
    self[dates, variants] = ...
    ```

    The `dates` is a `Dater` or a tuple of `Daters` or a time `Span` object,
    and `variants` is an integer or a tuple of integers or a `slice` object
    specifying the variants. The data assignment syntax sets the data in the
    time series.


    ### Time `Series` recreation ###

    ```
    self(dates)
    self(dates, variants)
    ```

    The `dates` is a `Dater` or a tuple of `Daters` or a time `Span` object,
    and `variants` is an integer or a tuple of integers or a `slice` object
    specifying the variants. The time `Series` recreation syntax returns a new
    time `Series` object based on the data selected by the `dates` and
    `variants`.

    ················································································
        """
        raise NotImplementedError

    def __getitem__(
        self,
        index: int | tuple,
    ) -> _np.ndarray:
        """
        Get data self[dates] or self[dates, variants]
        """
        if isinstance(index, int, ):
            # Time shift
            new = self.copy()
            new.shift(index, )
            return new
        else:
            # Extract data
            if isinstance(index, tuple, ):
                dates, variants = index
            else:
                dates, variants = index, None
            # dates = _dates.ensure_period_tuple(dates, frequency=self.frequency, )
            return self.get_data(dates, variants, )

    def __setitem__(
        self,
        index: int | tuple,
        data,
    ) -> None:
        """
        Set data self[dates] = ... or self[dates, variants] = ...
        """
        if not isinstance(index, tuple):
            index = (index, None, )
        return self.set_data(index[0], data, index[1], )

    def __call__(
        self,
        *index,
    ) -> Self:
        """
        Create a new time series based on date retrieved by self[dates] or self[dates, variants]
        """
        return self._get_data_and_recreate(*index, )

    #]




r"""
"""


from .dates import *
from .dates import __all__ as dates_all

from .series import *
from .series import __all__ as series_all

from .ez_plotly import *
from .ez_plotly import __all__ as ez_plotly_all

from .databoxes import *
from .databoxes import __all__ as databoxes_all

from .chartpacks.main import *
from .chartpacks.main import __all__ as chartpacks_all

__all__ = (
    *dates_all,
    *series_all,
    *ez_plotly_all,
    *databoxes_all,
    *chartpacks_all,
)




"""
"""


#[

from __future__ import annotations

import numpy as _np
import textwrap as _tw

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self
    from numbers import Real
    from . import Series, AxisType

#]


_NANABLE_FUNCTIONS = (
    "sum",
    "prod",
    "mean",
    "median",
    "std",
    "var",
    "max",
    "min",
    "percentile",
    "quantile",
)

_NAN_FUNCTIONS = tuple( f"nan{n}" for n in _NANABLE_FUNCTIONS )


_ALL_FUNCTIONS = _NANABLE_FUNCTIONS + _NAN_FUNCTIONS


#-------------------------------------------------------------------------------
# Mixin methods
#-------------------------------------------------------------------------------


_METHOD_TEMPLATE = """
    def {n}(
        self: Self,
        *args,
        **kwargs,
    ) -> None:
        r'''
        Function {n}
        '''
        num_periods = self.data.shape[0]
        self.data = _np.{n}(self.data, *args, axis=1, **kwargs, ).T.reshape(num_periods, -1, )
        self.trim()
    """


class Mixin:
    #[

    for n in _ALL_FUNCTIONS:
        code = _METHOD_TEMPLATE.format(n=n, )
        exec(_tw.dedent(code, ), )

    #]


#-------------------------------------------------------------------------------
# Functional forms
#-------------------------------------------------------------------------------


_FUNCTION_TEMPLATE = """
    def {n}(
        self: Series,
        *args,
        axis: AxisType = 1,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> Real | list[Real] | Series:
        if axis == 0:
            result = _np.{n}(self.data, *args, axis=0, **kwargs, ).T.tolist()
            if unpack_singleton and len(result) == 1:
                result = result[0]
            return result
        elif axis == 1:
            new = self.copy()
            new.{n}(*args, **kwargs, )
            return new
        else:
            raise ValueError(f'Axis must be 0 or 1; got {{axis}}')
"""


_functional_forms = set(_ALL_FUNCTIONS)
functional_form_context = {}

for n in _functional_forms:
    code = _FUNCTION_TEMPLATE.format(n=n, )
    exec(_tw.dedent(code, ), )
    functional_form_context[n] = locals()[n]


__all__ = tuple(_functional_forms)





r"""
Time series module
"""


from .main import *
from .main import __all__ as _main__all__

from .functions import *
from .functions import __all__ as _functions__all__

__all__ = (
    *_main__all__,
    *_functions__all__,
)

from .functions import functional_form_context




r"""
"""


#[

from __future__ import annotations

import textwrap as _tw

from ._functionalize import FUNC_STRING

#]


#-------------------------------------------------------------------------------
# Mixin methods
#-------------------------------------------------------------------------------


class Mixin:
    #[

    def shift(
        self,
        by: ShiftType = -1,
        **kwargs,
    ) -> None:
        r"""
................................................................................

==Shift the time series start date==

Shift the start date of the time series by a number of periods or to a specific
date.

    self.shift(by=-1, )

### Input arguments ###

???+ input "self"
    The current time series object that will be shifted.

???+ input "by"
    The number of periods to shift the observations by. If `by` is a string,
    the observations are manipulated as follows:

    * `"yoy"`: Shift all observations by one year back.

    * `"soy"`: Shift to the start of the year.

    * `"eopy"`: Shift to the end of the previous year.

................................................................................
        """
        if isinstance(by, int):
            self._shift_by_number(by, **kwargs, )
        else:
            method_name = f"_shift_{by}"
            getattr(self, method_name)(**kwargs, )

    def redate(
        self,
        new_date: Period,
        old_date: Period | None = None,
    ) -> None:
        r"""
        """
        if old_data is None:
            self.start = new_date
        else:
            self.start = new_date - (old_date - self.start)

    #]


#-------------------------------------------------------------------------------
# Functional forms
#-------------------------------------------------------------------------------


_functional_forms = {"shift", "redate", }

for n in _functional_forms:
    code = FUNC_STRING.format(n=n, )
    exec(_tw.dedent(code, ), globals(), )

__all__ = tuple(_functional_forms)




r"""
Description mixin
"""


#[

from __future__ import annotations

from typing import Protocol
import documark as _dm

#]


class HasDescriptionProtocol(Protocol, ):
    r"""
    """
    #[

    _description: str | None = None

    #]


#-------------------------------------------------------------------------------
# Mixin methods
#-------------------------------------------------------------------------------


class Mixin:
    #[

    @_dm.reference(category="information", )
    def get_description(
        self: HasDescriptionProtocol,
    ) -> str:
        r"""
................................................................................


==Get description attached to an object==

description = self.get_description()


### Input arguments ###

???+ input "self"
An object from which to get the description.


### Returns ###

???+ returns "description"
The description attached to the object.


................................................................................
        """
        return str(self._description or "")

    @_dm.reference(category="information", )
    def set_description(
            self: HasDescriptionProtocol,
            description: str,
        ) -> None:
        r"""
................................................................................

==Set the description for an object==

self.set_description(
    description,
)


### Input arguments ###

???+ input "self"
An Iris Pie object to which to attach the description.


???+ input "description"
The description to attach to the Iris Pie object.


### Returns ###

This method modifies the Iris Pie object in place and returns `None`.

................................................................................
        """
        self._description = str(description or "")

    #]


__all__ = {}




r"""
"""


#[

from __future__ import annotations

import documark as _dm

from ..dates import Period, Frequency

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Self, Callable

#]


#-------------------------------------------------------------------------------
# Mixin methods
#-------------------------------------------------------------------------------


class Mixin:
    #[

    @_dm.reference(category="serialization", )
    def to_jsonable(
        self,
        *,
        period_to_string: Callable = Period.to_sdmx_string,
        include_description: bool = True,
        include_frequency: bool = True,
        allow_multiple_variants: bool = False,
    ) -> dict:
        r"""
................................................................................

==Convert time series to a JSON-serializable dictionary==

This method converts the time series into a JSON-serializable dictionary format.
It includes options to include or exclude the description and frequency of the
series, as well as handling multiple variants.

    jsonable = self.to_jsonable(
        period_to_string=Period.to_sdmx_string,
        include_description=True,
        include_frequency=True,
        allow_multiple_variants=False,
    )


### Input arguments ###

???+ input "period_to_string"
    Function to convert a Period object to a string; choose from:
    * Period.to_sdmx_string (default): Converts to SDMX format (e.g., "2023-Q1").
    * Period.to_iso_string: Converts to ISO format (e.g., "2023-01-01").
    * custom function taking a Period and returning a string.

???+ input "include_description"
    If `True` (default), the `jsonable` dictionary includes the series
    description.

???+ input "include_frequency"
    If `True` (default), the `jsonable` dictionary includes the series
    frequency; this is recommended for clarity and robustness.

???+ input "allow_multiple_variants"
    Not implemented yet.


### Returns ###

???+ returns "jsonable"
    A dictionary representing the time series in a JSON-serializable format.

................................................................................
        """
        if allow_multiple_variants:
            raise NotImplementedError("allow_multiple_variants=True is not implemented yet.")
        #
        if self.num_variants > 1 and not allow_multiple_variants:
            raise ValueError("The series has multiple variants. Set allow_multiple_variants=True to serialize all variants.")
        #
        jsonable = {}
        #
        if include_description:
            jsonable["description"] = self.get_description() or ""
        #
        if include_frequency:
            jsonable["frequency"] = self.frequency.to_jsonable()
        #
        jsonable["start"] = (
            period_to_string(self.start)
            if self.start is not None else None
        )
        #
        jsonable["values"] = self.get_values(unpack_singleton=not allow_multiple_variants)
        #
        return jsonable

    @classmethod
    @_dm.reference(
        category="constructor",
        call_name="Series.from_jsonable",
    )
    def from_jsonable(
        klass,
        jsonable: dict,
        *,
        period_from_string: Callable = Period.from_sdmx_string,
        frequency_included: bool = True,
        description_included: bool = True,
        allow_multiple_variants: bool = False,
    ) -> Self:
        r"""
................................................................................

==Create time series from a JSON-serializable dictionary==

This class method creates a time series from a JSON-serializable dictionary.

    self = Series.from_jsonable(
        jsonable,
        period_from_string=Period.from_sdmx_string,
        frequency_included=True,
        description_included=True,
        allow_multiple_variants=False,
    )


### Input arguments ###

???+ input "jsonable"
    A dictionary representing the time series in a JSON-serializable format.

???+ input "period_from_string"
    Function to convert a string to a Period object; choose from:
    * Period.from_sdmx_string (default): Parses SDMX format (e.g., "2023-Q1").
    * Period.from_iso_string: Parses ISO format (e.g., "2023-01-01"), needs a `frequency` to be included for non-daily periods.
    * custom function taking a string and a frequency, and returning a Period.

???+ input "frequency_included"
    If `True` (default), the `jsonable` dictionary is expected to include the
    series frequency represented as a single letter.

???+ input "description_included"
    If `True` (default), the `jsonable` dictionary is expected to include the
    series description.

???+ input "allow_multiple_variants"
    Not implemented yet.


### Returns ###

???+ returns "self"
    A time series object created from the `jsonable` dictionary.

................................................................................
        """
        if allow_multiple_variants:
            raise NotImplementedError("allow_multiple_variants=True is not implemented yet.")
        #
        frequency = (
            Frequency.from_jsonable(jsonable["frequency"], )
            if frequency_included else None
        )
        #
        start = (
            period_from_string(jsonable["start"], frequency=frequency, )
            if jsonable["start"] else None
        )
        #
        values = jsonable["values"]
        if not allow_multiple_variants:
            values = tuple(values)
        #
        description = (
            jsonable["description"]
            if description_included else None
        )
        #
        if not start or not values:
            return klass.as_empty(description=description, )
        #
        return klass(
            start=start,
            values=values,
            description=description,
        )

        #]

